class InvalidConfigurationError(Exception):
    pass

class RootPathUndefinedError(InvalidConfigurationError):
    pass

class RootPathIsNotADirectoryError(InvalidConfigurationError):
    pass

class NonRenderableNodeError(Exception):
    pass

class ExportPathCollisionError(Exception):
    pass

class InvalidAttributeNameError(Exception):
    pass



import argparse

EXAMPLE1 = """# Example program: render a page without defining a site.
#
# Running this program prints a HTML document to standard output.
#
from ophinode import *

class MainPage:
    def body(self):
        return Div(
            H1("Main Page"),
            P("Welcome to ophinode!")
        )

    def head(self):
        return [
            Meta(charset="utf-8"),
            Title("Main Page")
        ]

print(render_page(MainPage(), HTML5Layout()))
"""

EXAMPLE2 = """# Example program: create a page in a directory.
#
# Running this program creates "index.html" in "./out" directory.
#
from ophinode import *

class DefaultLayout(Layout):
    def build(self, page, context):
        return [
            HTML5Doctype(),
            Html(
                Head(
                    Meta(charset="utf-8"),
                    Title(page.title()),
                    page.head()
                ),
                Body(
                    page.body()
                ),
            )
        ]

class MainPage:
    @property
    def layout(self):
        return DefaultLayout()

    def body(self):
        return Div(
            H1("Main Page"),
            P("Welcome to ophinode!")
        )

    def head(self):
        return []

    def title(self):
        return "Main Page"

if __name__ == "__main__":
    site = Site({
        "default_layout": DefaultLayout(),
        "export_root_path": "./out",
        "default_page_output_filename": "index.html",
    }, [
        ("/", MainPage()),
    ])

    site.build_site()
"""

def main():
    parser = argparse.ArgumentParser(prog="ophinode")
    parser.add_argument("subcommand", choices=["examples"])
    parser.add_argument("arguments", nargs="*")
    args = parser.parse_args()
    if args.subcommand == "examples":
        if not args.arguments:
            print("available examples: render_page, basic_site")
        elif args.arguments[0] == "render_page":
            print(EXAMPLE1)
        elif args.arguments[0] == "basic_site":
            print(EXAMPLE2)
        else:
            print("available examples: render_page, basic_site")
    else:
        parser.print_help()

if __name__ == "__main__":
    main()



from .base import *
from ophinode.exceptions import InvalidAttributeNameError

_default_escape_ampersands = False
_default_escape_tag_delimiters = True

def default_escape_ampersands(value: bool = True):
    global _default_escape_ampersands
    _default_escape_ampersands = bool(value)

def default_escape_tag_delimiters(value: bool = True):
    global _default_escape_tag_delimiters
    _default_escape_tag_delimiters = bool(value)

class HTML5Page(Page):
    @property
    def layout(self):
        return HTML5Layout()

    def html_attributes(self) -> dict:
        return {}

    def head(self):
        return []

    def body(self):
        return []

class HTML5Layout(Layout):
    def build(self, page: HTML5Page, context: "ophinode.rendering.RenderContext"):
        return [
            HTML5Doctype(),
            Html(
                Head(
                    page.head()
                ),
                Body(
                    page.body()
                ),
                **page.html_attributes()
            )
        ]

class Node:
    pass

class TextNode(Node, ClosedRenderable):
    def __init__(
        self,
        text_content: str,
        *,
        escape_ampersands: bool = None,
        escape_tag_delimiters: bool = None,
    ):
        self._text_content = text_content
        self._escape_ampersands = escape_ampersands
        self._escape_tag_delimiters = escape_tag_delimiters

    def render(self, context: "ophinode.rendering.RenderContext"):
        text_content = self._text_content

        escape_ampersands = self._escape_ampersands
        if escape_ampersands is None:
            escape_ampersands = _default_escape_ampersands
        if escape_ampersands:
            text_content = text_content.replace("&", "&amp;")

        escape_tag_delimiters = self._escape_tag_delimiters
        if escape_tag_delimiters is None:
            escape_tag_delimiters = _default_escape_tag_delimiters
        if escape_tag_delimiters:
            text_content = text_content.replace("<", "&lt;").replace(">", "&gt;")

        return text_content

    def escape_ampersands(self, value: bool = True):
        self._escape_ampersands = bool(value)
        return self

    def escape_tag_delimiters(self, value: bool = True):
        self._escape_tag_delimiters = bool(value)
        return self

class HTML5Doctype(Node, ClosedRenderable):
    def render(self, context: "ophinode.rendering.RenderContext"):
        return "<!doctype html>"

class CDATASection(Node, OpenRenderable, Expandable, Preparable):
    def __init__(self, *args):
        self._children = list(args)

    def prepare(self, context: "ophinode.rendering.RenderContext"):
        for c in self._children:
            if isinstance(c, Preparable):
                c.prepare(context)

    def expand(self, context: "ophinode.rendering.RenderContext"):
        return self._children.copy()

    def render_start(self, context: "ophinode.rendering.RenderContext"):
        return "<![CDATA[".format(self.tag)

    def render_end(self, context: "ophinode.rendering.RenderContext"):
        return "]]>".format(self.tag)

    @property
    def children(self):
        return self._children

    @property
    def auto_newline(self):
        return False

    @property
    def auto_indent(self):
        return False

class Comment(Node, OpenRenderable, Expandable, Preparable):
    def __init__(self, *args):
        self._children = list(args)

    def prepare(self, context: "ophinode.rendering.RenderContext"):
        for c in self._children:
            if isinstance(c, Preparable):
                c.prepare(context)

    def expand(self, context: "ophinode.rendering.RenderContext"):
        return self._children.copy()

    def render_start(self, context: "ophinode.rendering.RenderContext"):
        return "<!--".format(self.tag)

    def render_end(self, context: "ophinode.rendering.RenderContext"):
        return "-->".format(self.tag)

    @property
    def auto_newline(self):
        return False

    @property
    def auto_indent(self):
        return False

class Element(Node):
    def render_attributes(self):
        attribute_order = []
        keys = set(self._attributes)
        for k in ["id", "class", "style", "title"]:
            if k in keys:
                attribute_order.append(k)
                keys.remove(k)
        attribute_order += sorted(keys)

        rendered = []
        for k in attribute_order:
            for c in k:
                if c in " \"'>/=":
                    raise InvalidAttributeNameError(k)
            v = self._attributes[k]
            if isinstance(v, bool):
                if v:
                    rendered.append("{}".format(k))
            else:
                escaped = str(v)

                escape_ampersands = self._escape_ampersands
                if escape_ampersands is None:
                    escape_ampersands = _default_escape_ampersands
                if escape_ampersands:
                    escaped = escaped.replace("&", "&amp;")

                escape_tag_delimiters = self._escape_tag_delimiters
                if escape_tag_delimiters is None:
                    escape_tag_delimiters = _default_escape_tag_delimiters
                if escape_tag_delimiters:
                    escaped = escaped.replace("<", "&lt;").replace(">", "&gt;")

                escaped = escaped.replace("\"", "&quot;")
                rendered.append("{}=\"{}\"".format(k, escaped))

        return " ".join(rendered)

    @property
    def attributes(self):
        return self._attributes

    def escape_ampersands(self, value: bool = True):
        self._escape_ampersands = bool(value)
        return self

    def escape_tag_delimiters(self, value: bool = True):
        self._escape_tag_delimiters = bool(value)
        return self

class OpenElement(Element, OpenRenderable, Expandable, Preparable):
    tag = "div"
    render_mode = "hierarchy"

    def __init__(
        self,
        *args,
        cls = None,
        className = None,
        class_name = None,
        htmlFor = None,
        html_for = None,
        htmlAs = None,
        html_as = None,
        htmlAsync = None,
        html_async = None,
        accept_charset = None,
        escape_ampersands = None,
        escape_tag_delimiters = None,
        children = None,
        **kwargs
    ):
        self._children = list(args)
        if children is not None:
            for c in children:
                self._children.append(c)
        self._attributes = dict(kwargs)
        if cls is not None:
            self._attributes["class"] = cls
        if className is not None:
            self._attributes["class"] = className
        if class_name is not None:
            self._attributes["class"] = class_name
        if htmlFor is not None:
            self._attributes["for"] = htmlFor
        if html_for is not None:
            self._attributes["for"] = html_for
        if htmlAs is not None:
            self._attributes["as"] = htmlAs
        if html_as is not None:
            self._attributes["as"] = html_as
        if htmlAsync is not None:
            self._attributes["async"] = htmlAsync
        if html_async is not None:
            self._attributes["async"] = html_async
        if accept_charset is not None:
            self._attributes["accept-charset"] = accept_charset
        self._escape_ampersands = escape_ampersands
        self._escape_tag_delimiters = escape_tag_delimiters

    def prepare(self, context: "ophinode.rendering.RenderContext"):
        for c in self._children:
            if isinstance(c, Preparable):
                c.prepare(context)

    def expand(self, context: "ophinode.rendering.RenderContext"):
        expansion = []
        for c in self._children:
            if isinstance(c, str):
                node = TextNode(c)
                if self._escape_ampersands is not None:
                    node.escape_ampersands(self._escape_ampersands)
                if self._escape_tag_delimiters is not None:
                    node.escape_tag_delimiters(self._escape_tag_delimiters)
                expansion.append(node)
            else:
                expansion.append(c)
        return expansion

    def render_start(self, context: "ophinode.rendering.RenderContext"):
        rendered_attributes = self.render_attributes()
        if rendered_attributes:
            return "<{} {}>".format(self.tag, rendered_attributes)
        return "<{}>".format(self.tag)

    def render_end(self, context: "ophinode.rendering.RenderContext"):
        return "</{}>".format(self.tag)

    @property
    def children(self):
        return self._children

    @property
    def auto_newline(self):
        return self.render_mode == "hierarchy"

    @property
    def auto_indent(self):
        return self.render_mode == "phrase" or self.render_mode == "hierarchy"

class ClosedElement(Element, ClosedRenderable):
    tag = "meta"

    def __init__(
        self,
        *,
        cls = None,
        className = None,
        class_name = None,
        htmlFor = None,
        html_for = None,
        htmlAs = None,
        html_as = None,
        htmlAsync = None,
        html_async = None,
        accept_charset = None,
        escape_ampersands = None,
        escape_tag_delimiters = None,
        **kwargs
    ):
        self._attributes = dict(kwargs)
        if cls is not None:
            self._attributes["class"] = cls
        if className is not None:
            self._attributes["class"] = className
        if class_name is not None:
            self._attributes["class"] = class_name
        if htmlFor is not None:
            self._attributes["for"] = htmlFor
        if html_for is not None:
            self._attributes["for"] = html_for
        if htmlAs is not None:
            self._attributes["as"] = htmlAs
        if html_as is not None:
            self._attributes["as"] = html_as
        if htmlAsync is not None:
            self._attributes["async"] = htmlAsync
        if html_async is not None:
            self._attributes["async"] = html_async
        if accept_charset is not None:
            self._attributes["accept-charset"] = accept_charset
        self._escape_ampersands = escape_ampersands
        self._escape_tag_delimiters = escape_tag_delimiters

    def render(self, context: "ophinode.rendering.RenderContext"):
        rendered_attributes = self.render_attributes()
        if rendered_attributes:
            return "<{} {}>".format(self.tag, rendered_attributes)
        return "<{}>".format(self.tag)

# --- The document element ---

class HtmlElement(OpenElement):
    tag = "html"

Html = HtmlElement

# --- Document metadata ---

class HeadElement(OpenElement):
    tag = "head"

class TitleElement(OpenElement):
    tag = "title"
    render_mode = "inline"

class BaseElement(ClosedElement):
    tag = "base"

class LinkElement(ClosedElement):
    tag = "link"

class MetaElement(ClosedElement):
    tag = "meta"

class StyleElement(OpenElement):
    tag = "style"

    def __init__(self, *args, escape_tag_delimiters = None, **kwargs):
        if escape_tag_delimiters is None:
            # stylesheets might contain angle brackets, so it is better to
            # disable tag delimiter escaping by default
            escape_tag_delimiters = False
        super().__init__(
            *args,
            escape_tag_delimiters=escape_tag_delimiters,
            **kwargs
        )

    def expand(self, context: "ophinode.rendering.RenderContext"):
        expansion = []
        for c in self._children:
            if isinstance(c, str):
                # Stylesheets might contain "</style", so it must be escaped
                content = c.replace("</script", "\\3C/script")
                node = TextNode(content)
                if self._escape_ampersands is not None:
                    node.escape_ampersands(self._escape_ampersands)
                if self._escape_tag_delimiters is not None:
                    node.escape_tag_delimiters(self._escape_tag_delimiters)
                expansion.append(node)
            else:
                expansion.append(c)
        return expansion

Head = HeadElement
Title = TitleElement
Base = BaseElement
Link = LinkElement
Meta = MetaElement
Style = StyleElement

# --- Sections ---

class BodyElement(OpenElement):
    tag = "body"

class ArticleElement(OpenElement):
    tag = "article"

class SectionElement(OpenElement):
    tag = "section"

class NavigationElement(OpenElement):
    tag = "nav"

class AsideElement(OpenElement):
    tag = "aside"

class HeadingLevel1Element(OpenElement):
    tag = "h1"
    render_mode = "inline"

class HeadingLevel2Element(OpenElement):
    tag = "h2"
    render_mode = "inline"

class HeadingLevel3Element(OpenElement):
    tag = "h3"
    render_mode = "inline"

class HeadingLevel4Element(OpenElement):
    tag = "h4"
    render_mode = "inline"

class HeadingLevel5Element(OpenElement):
    tag = "h5"
    render_mode = "inline"

class HeadingLevel6Element(OpenElement):
    tag = "h6"
    render_mode = "inline"

class HeadingGroupElement(OpenElement):
    tag = "hgroup"

class HeaderElement(OpenElement):
    tag = "header"

class FooterElement(OpenElement):
    tag = "footer"

class AddressElement(OpenElement):
    tag = "address"

Body = BodyElement
Article = ArticleElement
Section = SectionElement
Nav = NavigationElement
Aside = AsideElement
H1 = HeadingLevel1Element
H2 = HeadingLevel2Element
H3 = HeadingLevel3Element
H4 = HeadingLevel4Element
H5 = HeadingLevel5Element
H6 = HeadingLevel6Element
HGroup = HeadingGroupElement
Hgroup = HeadingGroupElement
Header = HeaderElement
Footer = FooterElement
Address = AddressElement

# --- Grouping content ---

class ParagraphElement(OpenElement):
    tag = "p"
    render_mode = "phrase"

class HorizontalRuleElement(ClosedElement):
    tag = "hr"

class PreformattedTextElement(OpenElement):
    tag = "pre"
    render_mode = "pre"

class BlockQuotationElement(OpenElement):
    tag = "blockquote"

class OrderedListElement(OpenElement):
    tag = "ol"

class UnorderedListElement(OpenElement):
    tag = "ul"

class MenuElement(OpenElement):
    tag = "menu"

class ListItemElement(OpenElement):
    tag = "li"

class DescriptionListElement(OpenElement):
    tag = "dl"

class DescriptionTermElement(OpenElement):
    tag = "dt"
    render_mode = "inline"

class DescriptionDetailsElement(OpenElement):
    tag = "dd"
    render_mode = "inline"

class FigureElement(OpenElement):
    tag = "figure"

class FigureCaptionElement(OpenElement):
    tag = "figcaption"

class MainElement(OpenElement):
    tag = "main"

class SearchElement(OpenElement):
    tag = "search"

class DivisionElement(OpenElement):
    tag = "div"

P = ParagraphElement
HR = HorizontalRuleElement
Hr = HorizontalRuleElement
Pre = PreformattedTextElement
BlockQuote = BlockQuotationElement
Blockquote = BlockQuotationElement
OL = OrderedListElement
Ol = OrderedListElement
UL = UnorderedListElement
Ul = UnorderedListElement
Menu = MenuElement
LI = ListItemElement
Li = ListItemElement
DL = DescriptionListElement
Dl = DescriptionListElement
DT = DescriptionTermElement
Dt = DescriptionTermElement
DD = DescriptionDetailsElement
Dd = DescriptionDetailsElement
Figure = FigureElement
FigCaption = FigureCaptionElement
Figcaption = FigureCaptionElement
Main = MainElement
Search = SearchElement
Div = DivisionElement

# --- Text-level semantics ---

class AnchorElement(OpenElement):
    tag = "a"

class EmphasisElement(OpenElement):
    tag = "em"
    render_mode = "inline"

class StrongImportanceElement(OpenElement):
    tag = "strong"
    render_mode = "inline"

class SmallPrintElement(OpenElement):
    tag = "small"
    render_mode = "inline"

class StrikethroughElement(OpenElement):
    tag = "s"
    render_mode = "inline"

class CitationElement(OpenElement):
    tag = "cite"
    render_mode = "inline"

class QuotationElement(OpenElement):
    tag = "q"
    render_mode = "inline"

class DefinitionElement(OpenElement):
    tag = "dfn"
    render_mode = "inline"

class AbbreviationElement(OpenElement):
    tag = "abbr"
    render_mode = "inline"

class RubyAnnotationElement(OpenElement):
    tag = "ruby"
    render_mode = "inline"

class RubyTextElement(OpenElement):
    tag = "rt"
    render_mode = "inline"

class RubyParenthesesElement(OpenElement):
    tag = "rp"
    render_mode = "inline"

class DataElement(OpenElement):
    tag = "data"
    render_mode = "inline"

class TimeElement(OpenElement):
    tag = "time"
    render_mode = "inline"

class CodeElement(OpenElement):
    tag = "code"
    render_mode = "inline"

class VariableElement(OpenElement):
    tag = "var"
    render_mode = "inline"

class SampleElement(OpenElement):
    tag = "samp"
    render_mode = "inline"

class KeyboardInputElement(OpenElement):
    tag = "kbd"
    render_mode = "inline"

class SubscriptElement(OpenElement):
    tag = "sub"
    render_mode = "inline"

class SuperscriptElement(OpenElement):
    tag = "sup"
    render_mode = "inline"

class ItalicTextElement(OpenElement):
    tag = "i"
    render_mode = "inline"

class BoldTextElement(OpenElement):
    tag = "b"
    render_mode = "inline"

class UnarticulatedAnnotationElement(OpenElement):
    tag = "u"
    render_mode = "inline"

class MarkedTextElement(OpenElement):
    tag = "mark"
    render_mode = "inline"

class BidirectionalIsolateElement(OpenElement):
    tag = "bdi"
    render_mode = "inline"

class BidirectionalOverrideElement(OpenElement):
    tag = "bdo"
    render_mode = "inline"

class SpanElement(OpenElement):
    tag = "span"
    render_mode = "inline"

class LineBreakElement(ClosedElement):
    tag = "br"

class LineBreakOpportunityElement(ClosedElement):
    tag = "wbr"

A = AnchorElement
EM = EmphasisElement
Em = EmphasisElement
Strong = StrongImportanceElement
Small = SmallPrintElement
S = StrikethroughElement
Cite = CitationElement
Q = QuotationElement
Dfn = DefinitionElement
Abbr = AbbreviationElement
Ruby = RubyAnnotationElement
RT = RubyTextElement
Rt = RubyTextElement
RP = RubyParenthesesElement
Rp = RubyParenthesesElement
Data = DataElement
Time = TimeElement
Code = CodeElement
Var = VariableElement
Samp = SampleElement
Kbd = KeyboardInputElement
Sub = SubscriptElement
Sup = SuperscriptElement
I = ItalicTextElement
B = BoldTextElement
U = UnarticulatedAnnotationElement
Mark = MarkedTextElement
BDI = BidirectionalIsolateElement
Bdi = BidirectionalIsolateElement
BDO = BidirectionalOverrideElement
Bdo = BidirectionalOverrideElement
Span = SpanElement
BR = LineBreakElement
Br = LineBreakElement
WBR = LineBreakOpportunityElement
Wbr = LineBreakOpportunityElement

# --- Edits ---

class InsertionElement(OpenElement):
    tag = "ins"

class DeletionElement(OpenElement):
    tag = "del"
    render_mode = "inline"

Ins = InsertionElement
Del = DeletionElement

# --- Embedded content ---

class PictureElement(OpenElement):
    tag = "picture"

class SourceElement(ClosedElement):
    tag = "source"

class ImageElement(ClosedElement):
    tag = "image"

class InlineFrameElement(OpenElement):
    tag = "iframe"

class EmbeddedContentElement(ClosedElement):
    tag = "embed"

class ExternalObjectElement(OpenElement):
    tag = "object"

class VideoElement(OpenElement):
    tag = "video"

class AudioElement(OpenElement):
    tag = "audio"

class TextTrackElement(ClosedElement):
    tag = "track"

class ImageMapElement(OpenElement):
    tag = "map"

class ImageMapAreaElement(ClosedElement):
    tag = "area"

Picture = PictureElement
Source = SourceElement
Img = ImageElement
IFrame = InlineFrameElement
Iframe = InlineFrameElement
Embed = EmbeddedContentElement
Object = ExternalObjectElement
Video = VideoElement
Audio = AudioElement
Track = TextTrackElement
Map = ImageMapElement
Area = ImageMapAreaElement

# --- Tabular data ---

class TableElement(OpenElement):
    tag = "table"

class TableCaptionElement(OpenElement):
    tag = "caption"

class TableColumnGroupElement(OpenElement):
    tag = "colgroup"

class TableColumnElement(ClosedElement):
    tag = "col"

class TableBodyElement(OpenElement):
    tag = "tbody"

class TableHeadElement(OpenElement):
    tag = "thead"

class TableFootElement(OpenElement):
    tag = "tfoot"

class TableRowElement(OpenElement):
    tag = "tr"

class TableDataCellElement(OpenElement):
    tag = "td"

class TableHeaderCellElement(OpenElement):
    tag = "th"

Table = TableElement
Caption = TableCaptionElement
ColGroup = TableColumnGroupElement
Colgroup = TableColumnGroupElement
Column = TableColumnElement
TBody = TableBodyElement
Tbody = TableBodyElement
THead = TableHeadElement
Thead = TableHeadElement
TFoot = TableFootElement
Tfoot = TableFootElement
TR = TableRowElement
Tr = TableRowElement
TD = TableDataCellElement
Td = TableDataCellElement
TH = TableHeaderCellElement
Th = TableHeaderCellElement

# --- Forms ---

class FormElement(OpenElement):
    tag = "form"

class LabelElement(OpenElement):
    tag = "label"

class InputElement(ClosedElement):
    tag = "input"

class ButtonElement(OpenElement):
    tag = "button"

class SelectElement(OpenElement):
    tag = "select"

class DataListElement(OpenElement):
    tag = "datalist"

class OptionGroupElement(OpenElement):
    tag = "optgroup"

class OptionElement(OpenElement):
    tag = "option"

class TextAreaElement(OpenElement):
    tag = "textarea"
    render_mode = "pre"

class OutputElement(OpenElement):
    tag = "output"

class ProgressElement(OpenElement):
    tag = "progress"
    render_mode = "inline"

class MeterElement(OpenElement):
    tag = "meter"
    render_mode = "inline"

class FieldSetElement(OpenElement):
    tag = "fieldset"

class FieldSetLegendElement(OpenElement):
    tag = "legend"
    render_mode = "inline"

Form = FormElement
Label = LabelElement
Input = InputElement
Button = ButtonElement
Select = SelectElement
DataList = DataListElement
Datalist = DataListElement
OptGroup = OptionGroupElement
Optgroup = OptionGroupElement
Option = OptionElement
Progress = ProgressElement
Meter = MeterElement
FieldSet = FieldSetElement
Fieldset = FieldSetElement
Legend = FieldSetLegendElement

# --- Interactive elements ---

class DetailsElement(OpenElement):
    tag = "details"

class SummaryElement(OpenElement):
    tag = "summary"
    render_mode = "inline"

class DialogElement(OpenElement):
    tag = "dialog"

Details = DetailsElement
Summary = SummaryElement
Dialog = DialogElement

# --- Scripting ---

class ScriptElement(OpenElement):
    tag = "script"

    def __init__(self, *args, escape_tag_delimiters = None, **kwargs):
        if escape_tag_delimiters is None:
            # javascript code might contain angle brackets,
            # so it is better to disable tag delimiter escaping by default
            escape_tag_delimiters = False
        super().__init__(
            *args,
            escape_tag_delimiters=escape_tag_delimiters,
            **kwargs
        )

    def expand(self, context: "ophinode.rendering.RenderContext"):
        expansion = []
        for c in self._children:
            if isinstance(c, str):
                # Due to restrictions for contents of script elements, some
                # sequences of characters must be replaced before constructing
                # a script element.
                # 
                # Unfortunately, correctly replacing such character sequences
                # require a full lexical analysis on the script content, but
                # ophinode is currently incapable of doing so.
                #
                # However, the sequences are expected to be rarely seen
                # outside literals, so replacements are done nonetheless.
                #
                # This behavior might change in the later versions of ophinode
                # when it starts to better support inline scripting.
                #
                # Read https://html.spec.whatwg.org/multipage/scripting.html#restrictions-for-contents-of-script-elements
                # for more information.
                #
                content = c.replace(
                    "<!--", "\\x3C!--"
                ).replace(
                    "<script", "\\x3Cscript"
                ).replace(
                    "</script", "\\x3C/script"
                )
                node = TextNode(content)
                if self._escape_ampersands is not None:
                    node.escape_ampersands(self._escape_ampersands)
                if self._escape_tag_delimiters is not None:
                    node.escape_tag_delimiters(self._escape_tag_delimiters)
                expansion.append(node)
            else:
                expansion.append(c)
        return expansion

class NoScriptElement(OpenElement):
    tag = "noscript"

class TemplateElement(OpenElement):
    tag = "template"

class SlotElement(OpenElement):
    tag = "slot"

class CanvasElement(OpenElement):
    tag = "canvas"

Script = ScriptElement
NoScript = NoScriptElement
Noscript = NoScriptElement
Template = TemplateElement
Slot = SlotElement
Canvas = CanvasElement




"Collection of various constants."

# Site options
EXPORT_ROOT_PATH_OPTION_KEY = "export_root_path"
EXPORT_ROOT_PATH_OPTION_DEFAULT_VALUE = ""
DEFAULT_LAYOUT_OPTION_KEY = "default_layout"
DEFAULT_LAYOUT_OPTION_DEFAULT_VALUE = None
DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_KEY = "default_page_output_filename"
DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_DEFAULT_VALUE = "index.html"
PAGE_OUTPUT_FILE_EXTENSION_OPTION_KEY = "page_output_file_extension"
PAGE_OUTPUT_FILE_EXTENSION_OPTION_DEFAULT_VALUE = "html"
AUTO_EXPORT_FILES_OPTION_KEY = "auto_export_files"
AUTO_EXPORT_FILES_OPTION_DEFAULT_VALUE = True



import json
import os.path
import pathlib
import collections
import collections.abc
from abc import ABC, abstractmethod
from typing import Any, Union, Iterable, Tuple, Callable

from .exceptions import *
from .constants import *
from .nodes.base import *
from .nodes.html import TextNode, HTML5Layout
from .rendering import RenderContext, RenderNode

class _StackDelimiter:
    pass

class Site:
    def __init__(
        self,
        options: Union[collections.abc.Mapping, None] = None,
        pages: Union[Iterable[Tuple[str, Any]], None] = None,
        processors: Union[
            Iterable[Tuple[str, Callable[["RenderContext"], None]]],
            None
        ] = None,
    ):
        self._options = {}
        if options is not None:
            self.update_options(options)

        self._pages_dict = {}
        self._pages = []
        if pages is not None:
            if not isinstance(pages, collections.abc.Iterable):
                raise TypeError("pages must be an iterable")
            for path, page in pages:
                if not isinstance(path, str):
                    raise ValueError("path to a page must be a str")
                if path in self._pages_dict:
                    raise ValueError("duplicate page path: " + path)
                self._pages_dict[path] = page
                self._pages.append((path, page))

        self._preprocessors_before_site_build = []
        self._postprocessors_after_site_build = []
        self._preprocessors_before_page_build_stage = []
        self._postprocessors_after_page_build_stage = []
        self._preprocessors_before_page_preparation_stage = []
        self._postprocessors_after_page_preparation_stage = []
        self._preprocessors_before_page_expansion_stage = []
        self._postprocessors_after_page_expansion_stage = []
        self._preprocessors_before_page_rendering_stage = []
        self._postprocessors_after_page_rendering_stage = []
        if processors is not None:
            if not isinstance(processors, collections.abc.Iterable):
                raise TypeError("processors must be an iterable")
            for stage, processor in processors:
                self.add_processor(stage, processor)

    def update_options(
        self,
        options: collections.abc.Mapping,
        ignore_invalid_keys: bool = False
    ):
        if not isinstance(options, collections.abc.Mapping):
            raise TypeError("options must be a mapping")
        for k, v in options.items():
            if k == EXPORT_ROOT_PATH_OPTION_KEY:
                self.set_export_root_path(v)
            elif k == DEFAULT_LAYOUT_OPTION_KEY:
                self.set_default_layout(v)
            elif k == DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_KEY:
                self.set_default_page_output_filename(v)
            elif k == PAGE_OUTPUT_FILE_EXTENSION_OPTION_KEY:
                self.set_page_output_file_extension(v)
            elif k == AUTO_EXPORT_FILES_OPTION_KEY:
                self.set_auto_export_files(v)
            elif not ignore_invalid_keys:
                raise ValueError("unknown option key: {}".format(k))

    def set_export_root_path(self, export_root_path: str):
        if not isinstance(export_root_path, str):
            raise TypeError("export_root_path must be a str")
        self._options[EXPORT_ROOT_PATH_OPTION_KEY] = os.path.abspath(
            export_root_path
        )

    @property
    def export_root_path(self) -> str:
        return self._options.get(
            EXPORT_ROOT_PATH_OPTION_KEY,
            EXPORT_ROOT_PATH_OPTION_DEFAULT_VALUE
        )

    def set_default_layout(self, default_layout: Layout):
        if not isinstance(default_layout, Layout):
            raise TypeError("default_layout must be a Layout")
        self._options[DEFAULT_LAYOUT_OPTION_KEY] = default_layout

    @property
    def default_layout(self) -> Union[Layout, None]:
        return self._options.get(
            DEFAULT_LAYOUT_OPTION_KEY,
            DEFAULT_LAYOUT_OPTION_DEFAULT_VALUE
        )

    def set_default_page_output_filename(
        self,
        default_page_output_filename: str
    ):
        if not isinstance(default_page_output_filename, str):
            raise TypeError("default_page_output_filename must be a str")
        self._options[
            DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_KEY
        ] = default_page_output_filename

    @property
    def default_page_output_filename(self) -> str:
        return self._options.get(
            DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_KEY,
            DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_DEFAULT_VALUE
        )

    def set_page_output_file_extension(
        self,
        page_output_file_extension: str
    ):
        if not isinstance(page_output_file_extension, str):
            raise TypeError("page_output_file_extension must be a str")
        self._options[
            PAGE_OUTPUT_FILE_EXTENSION_OPTION_KEY
        ] = page_output_file_extension

    @property
    def page_output_file_extension(self) -> str:
        return self._options.get(
            PAGE_OUTPUT_FILE_EXTENSION_OPTION_KEY,
            PAGE_OUTPUT_FILE_EXTENSION_OPTION_DEFAULT_VALUE
        )

    def set_auto_export_files(self, auto_export_files: bool):
        self._options[AUTO_EXPORT_FILES_OPTION_KEY] = bool(auto_export_files)

    @property
    def auto_export_files(self) -> bool:
        return self._options.get(
            AUTO_EXPORT_FILES_OPTION_KEY,
            AUTO_EXPORT_FILES_OPTION_DEFAULT_VALUE
        )

    def add_processor(
        self,
        stage: str,
        processor: Callable[["RenderContext"], None]
    ):
        if not isinstance(stage, str):
            raise ValueError("processor stage must be a str")
        if not callable(processor):
            raise TypeError("processor must be a callable")
        if stage == "pre_site_build":
            self._preprocessors_before_site_build.append(processor)
        elif stage == "post_site_build":
            self._postprocessors_after_site_build.append(processor)
        elif stage == "pre_page_build":
            self._preprocessors_before_page_build_stage.append(processor)
        elif stage == "post_page_build":
            self._postprocessors_after_page_build_stage.append(processor)
        elif stage == "pre_page_preparation":
            self._preprocessors_before_page_preparation_stage.append(processor)
        elif stage == "post_page_preparation":
            self._postprocessors_after_page_preparation_stage.append(processor)
        elif stage == "pre_page_expansion":
            self._preprocessors_before_page_expansion_stage.append(processor)
        elif stage == "post_page_expansion":
            self._postprocessors_after_page_expansion_stage.append(processor)
        elif stage == "pre_page_rendering":
            self._preprocessors_before_page_rendering_stage.append(processor)
        elif stage == "post_page_rendering":
            self._postprocessors_after_page_rendering_stage.append(processor)
        else:
            raise ValueError("invalid rendering stage: '{}'".format(stage))

    def _prepare_site(self) -> RenderContext:
        context = RenderContext(self)
        for path, page in self._pages:
            context._page_data[path] = {}
        for processor in self._preprocessors_before_site_build:
            processor(context)
        return context

    def _build_nodes(self, page: Any, context: RenderContext) -> Iterable:
        layout = None
        if hasattr(page, "layout"):
            layout = page.layout
            l_src = "layout property of page"
        if not layout:
            layout = self.default_layout
            l_src = "default_layout option of renderer"
        if not layout:
            layout = HTML5Layout()
            l_src = "fallback layout (HTML5Layout)"

        if callable(layout):
            layout = layout(context)

        if not isinstance(layout, Layout):
            raise ValueError(
                "resolved layout (from {}) is not a Layout instance".format(
                    l_src
                )
            )

        return layout.build(page, context)

    def _build_page(self, path, page, context):
        context._current_page_path = path
        context._current_page = page
        context.built_pages[path] = self._build_nodes(page, context)
        context._current_page = None
        context._current_page_path = None

    def _build_pages(self, context):
        for processor in self._preprocessors_before_page_build_stage:
            processor(context)
        for path, page in self._pages:
            self._build_page(path, page, context)
        for processor in self._postprocessors_after_page_build_stage:
            processor(context)

    def _prepare_nodes(self, page_built: Iterable, context: RenderContext):
        for node in page_built:
            if isinstance(node, Preparable):
                node.prepare(context)

    def _prepare_page(self, path: str, page: Any, context: RenderContext):
        context._current_page_path = path
        context._current_page = page
        self._prepare_nodes(context.built_pages[path], context)
        context._current_page = None
        context._current_page_path = None

    def _prepare_pages(self, context: RenderContext):
        for processor in self._preprocessors_before_page_preparation_stage:
            processor(context)
        for path, page in self._pages:
            self._prepare_page(path, page, context)
        for processor in self._postprocessors_after_page_preparation_stage:
            processor(context)

    def _expand_nodes(
        self,
        page_built: Iterable,
        context: RenderContext
    ) -> RenderNode:
        root_node = RenderNode(None)
        curr = root_node

        stack = collections.deque()
        for node in reversed(page_built):
            stack.append(node)

        while stack:
            node = stack.pop()
            if isinstance(node, _StackDelimiter):
                curr = curr._parent
            elif isinstance(node, str):
                render_node = RenderNode(TextNode(node))
                render_node._parent = curr
                curr._children.append(render_node)
            elif callable(node):
                r = node(context)
                stack.append(r)
            elif isinstance(node, collections.abc.Iterable):
                for n in reversed(node):
                    stack.append(n)
            elif isinstance(node, Expandable):
                r = node.expand(context)
                stack.append(_StackDelimiter())
                next_render_node = RenderNode(node)
                next_render_node._parent = curr
                curr._children.append(next_render_node)
                curr = next_render_node
                stack.append(r)
            else:
                next_render_node = RenderNode(node)
                next_render_node._parent = curr
                curr._children.append(next_render_node)

        return root_node

    def _expand_page(self, path: str, page: Any, context: RenderContext):
        context._current_page_path = path
        context._current_page = page
        root_node = self._expand_nodes(context.built_pages[path], context)
        context.expanded_pages[path] = root_node
        context._current_page = None
        context._current_page_path = None

    def _expand_pages(self, context: RenderContext):
        for processor in self._preprocessors_before_page_expansion_stage:
            processor(context)
        for path, page in self._pages:
            self._expand_page(path, page, context)
        for processor in self._postprocessors_after_page_expansion_stage:
            processor(context)

    def _render_page(self, path: str, page: Any, context: RenderContext):
        context._current_page_path = path
        context._current_page = page
        root_node = context.expanded_pages[path]
        render_result = root_node.render(context)
        context.rendered_pages[path] = render_result
        context._current_page = None
        context._current_page_path = None

    def _render_pages(self, context: RenderContext):
        for processor in self._preprocessors_before_page_rendering_stage:
            processor(context)
        for path, page in self._pages:
            self._render_page(path, page, context)
        for processor in self._postprocessors_after_page_rendering_stage:
            processor(context)

    def _finalize_site(self, context: RenderContext):
        for path, render_result in context.rendered_pages.items():
            page_path = path
            if path.endswith("/"):
                page_path += self.default_page_output_filename
            elif self.page_output_file_extension:
                fname = path[path.rfind("/")+1:]
                dot_index = fname.rfind(".")
                if (
                    (
                        dot_index == -1
                        or fname[dot_index+1:]
                           != self.page_output_file_extension
                    )
                    and
                    # the rightmost dot is not the first (or last) character
                    (
                        dot_index != 0
                        and (fname and dot_index != len(fname)-1)
                    )
                ):
                    page_path += "." + self.page_output_file_extension
            page_path = os.path.normpath(page_path)
            if page_path in context.exported_files:
                raise ExportPathCollisionError(
                    "attempted to export a page to '{}', but another file is "
                    "already exported to that path".format(page_path)
                )
            context.exported_files[page_path] = render_result

        for processor in self._postprocessors_after_site_build:
            processor(context)

    def _export_files(self, context: RenderContext):
        # Check if export_root_path has been set
        if not self.export_root_path:
            raise RootPathUndefinedError(
                "failed to export files because export_root_path is empty"
            )

        # Ensure export_root_path is a directory,
        # or if it does not exist, create one
        export_root_path = pathlib.Path(self.export_root_path)
        if not export_root_path.exists():
            if export_root_path.is_symlink():
                raise RootPathIsNotADirectoryError(
                    "failed to export files because export_root_path is a "
                    "broken symlink"
                )
            try:
                export_root_path.mkdir(parents=True)
            except NotADirectoryError as exc:
                raise RootPathIsNotADirectoryError(
                    "failed to export files because the parent of "
                    "export_root_path is not a directory, and thus "
                    "export_root_path cannot be a directory"
                ) from exc
        elif not export_root_path.is_dir():
            raise RootPathIsNotADirectoryError(
                "failed to export files because export_root_path is not a "
                "directory"
            )

        # Export files
        for path, file_content in context.exported_files.items():
            target_path = (
                pathlib.Path(self.export_root_path) / path.lstrip('/')
            )
            target_directory = target_path.parent
            if not target_directory.exists():
                target_directory.mkdir(parents=True)
            if isinstance(file_content, (bytes, bytearray)):
                with target_path.open(mode="wb") as f:
                    f.write(file_content)
            elif isinstance(file_content, str):
                with target_path.open(mode="w", encoding="utf-8") as f:
                    f.write(file_content)
            else:
                with target_path.open(mode="w", encoding="utf-8") as f:
                    json.dump(file_content, f, indent=2)

    def build_site(self):
        context = self._prepare_site()
        self._build_pages(context)
        self._prepare_pages(context)
        self._expand_pages(context)
        self._render_pages(context)
        self._finalize_site(context)
        if self.auto_export_files:
            self._export_files(context)
        return context

def render_page(page: Any, default_layout: Union[Layout, None] = None):
    options = {
        EXPORT_ROOT_PATH_OPTION_KEY: "/",
        AUTO_EXPORT_FILES_OPTION_KEY: False,
    }
    if default_layout is not None:
        options[DEFAULT_LAYOUT_OPTION_KEY] = default_layout
    site = Site(options, [("/", page)])
    context = site.build_site()
    return context.rendered_pages["/"]



from typing import Any
from abc import ABC, abstractmethod

class ClosedRenderable(ABC):
    @abstractmethod
    def render(self, context: "ophinode.rendering.RenderContext"):
        pass

class OpenRenderable(ABC):
    @abstractmethod
    def render_start(self, context: "ophinode.rendering.RenderContext"):
        pass

    @abstractmethod
    def render_end(self, context: "ophinode.rendering.RenderContext"):
        pass

    @property
    def auto_newline(self):
        return False

    @property
    def auto_indent(self):
        return False

class Expandable(ABC):
    @abstractmethod
    def expand(self, context: "ophinode.rendering.RenderContext"):
        pass

class Preparable(ABC):
    @abstractmethod
    def prepare(self, context: "ophinode.rendering.RenderContext"):
        pass

class Page:
    @property
    def layout(self):
        return None

class Layout(ABC):
    @abstractmethod
    def build(self, page: Any, context: "ophinode.rendering.RenderContext"):
        pass

